"""
  PyMusique - iTMS client

  Travis Watkins <alleykat@gmail.com>
  Cody Brocious <daeken@gmail.com>
  Jon Lech Johansen <jon@nanocrew.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  (C) Copyright 2005 Travis Watkins, Cody Brocious, Jon Lech Johansen
"""

import os
import sys
import debug

usegconf = False
usewinreg = False

musique_version = '0.5'
debug.verbosity = debug.ERROR

if not sys.platform == 'win32':
    try:
        import gconf
        usegconf = True
    except ImportError:
        pass
    homedir = os.path.expanduser('~')
    songdir = os.path.join(homedir, 'Music')
else:
    try:
        from winreg import *
        usewinreg = True
    except ImportError:
        pass
    #try:
    #    from win32com.shell import shell, shellcon
    #    homedir = shell.SHGetPathFromIDList(shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_APPDATA))
    #    songdir = shell.SHGetPathFromIDList(shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_MYMUSIC))
    #except ImportError:
    homedir = os.environ['APPDATA']
    songdir = os.path.join(os.environ['USERPROFILE'], 'Music')

drmsdir = os.path.join(homedir, '.drms')
cachedir = os.path.join(homedir, '.pymusique')

class Configuration:
    def __init__(self):
        global usegconf, usewinreg, songdir
        if usegconf:
            self.client = gconf.client_get_default()
            self.path = '/apps/pymusique'
            if not self.client.dir_exists(self.path):
                self.client.add_dir(self.path, gconf.CLIENT_PRELOAD_RECURSIVE)
        elif usewinreg:
            try:
                self.key = Key(HKCU, 'SOFTWARE\\PyMusique')
            except KeyNotFound:
                self.key = Key(HKCU, 'SOFTWARE').add('PyMusique')
        else:
            self.store = {}

        if self['songdir'] == None:
            self['songdir'] = songdir

    def __getitem__(self, key):
        global usegconf, usewinreg
        if usegconf:
            value = self.client.get(self.path + '/' + key)
            if (value):
                dtype = value.type
                if (dtype == gconf.VALUE_STRING):
                    return value.get_string()
                elif (dtype == gconf.VALUE_FLOAT):
                    return value.get_float()
                elif (dtype == gconf.VALUE_INT):
                    return value.get_int()
                elif (dtype == gconf.VALUE_BOOL):
                    return value.get_bool()
                else:
                    raise ValueError("Invalid data type %s." % (dtype))
            else:
                return None
        elif usewinreg:
            try:
                return self.key.values[key].value
            except ValueNotFound:
                return None
        else:
            try:
                return self.store[key]
            except KeyError:
                return None

    def __setitem__(self, key, value):
        global usegconf, usewinreg
        if usegconf:
            if (type(value) == str or type(value) == unicode):
                v = gconf.Value(gconf.VALUE_STRING)
                v.set_string(value)
            elif (type(value) == float):
                v = gconf.Value(gconf.VALUE_FLOAT)
                v.set_float(value)
            elif (type(value) == int):
                v = gconf.Value(gconf.VALUE_INT)
                v.set_int(value)
            elif (type(value) == bool):
                v = gconf.Value(gconf.VALUE_BOOL)
                v.set_bool(value)
            else:
                raise ValueError("Invalid data type %s." % (`type(value)`))

            self.client.set(self.path + '/' + key, v)
            return True
        elif usewinreg:
            try:
                self.key.set(key, value)
                return True
            except:
                return False
        else:
            self.store[key] = value
            return True
