"""
  pyMusique - iTMS client

  Travis Watkins <alleykat@gmail.com>
  Cody Brocious <daeken@gmail.com>
  Charles M. Hannum <root@ihack.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  (C) Copyright 2005 Travis Watkins, Cody Brocious, Charles M. Hannum
"""

import sys
import math
import struct

def uint32(data):
    if type(data) == str: return struct.unpack('>L', data[:4])[0]
    else: return struct.pack('>L', data)

def uint64(data):
    if type(data) == str: return struct.unpack('>Q', data[:8])[0]
    else: return struct.pack('>Q', data)

class QTStream:
    sbuffer = None
    tree = None
    containers = None
    special = None
    i = 0

    def __init__(self, f, containers, special):
        self.sbuffer = f
        self.containers, self.special = containers, special
        self.tree = self.getAtomTree()

    def printTree(self, tree=None):
        if not tree:self.tabs=1;tree=self.tree
        for x in tree:
            if x[3]: continue
            if self.tabs:
                print '  '*(self.tabs-1) + str(self.tabs) + ' |_ ' + x[1] + ' (' + str(len(x[2])) + ')'
            if type(x[2]) == list:
                self.tabs+=1
                self.printTree(x[2])
                self.tabs-=1
        return self

    def getData(self, tree=None, justlen=0):
        if not tree: tl=1;tree=self.tree
        else: tl=0
        s = ''
        for x in tree:
            if x[3] == 1: continue
            if x[1] in self.special:
                d = self.special[x[1]](self, x[2], 1 + justlen)
                if justlen:
                    d = ' '*d
            elif type(x[2]) == list:
                d = self.getData(x[2], justlen=justlen)
            else:
                d = x[2]
            if len(x[2]) + 8 > math.pow(2, 32):
                s += uint32(0) + x[1] + uint64(len(d) + 16) + d
            else:
                s += uint32(len(d) + 8) + x[1] + d
        return s

    def getAtom(self, atom, tree=None):
        if not tree: tl=1;tree=self.tree
        else: tl=0
        found = []
        for x in tree:
            if x[3] == 1: continue
            if x[1] == atom:
                found.append([x[0], x[2]])
            elif type(x[2]) == list:
                found += self.getAtom(atom, x[2])
        return found

    def setAtom(self, atomid, atom, value, tree=None):
        if not tree: tree=self.tree
        for x in tree:
            if x[0] == atomid:
                x[1] = atom
                x[2] = value
                return True
            if type(x[2]) == list:
                if self.setAtom(atomid, atom, value, x[2]):
                    return True
        return False

    def getAtomPos(self, atom, tree=None, l=0):
        if not tree: tl=1;tree=self.tree
        else: tl=0
        found = []
        for x in tree:
            if x[3] == 1: continue
            l += 8
            if x[1] == atom:
                found.append(l)
            if x[1] in self.special:
                l += self.special[x[1]](self, x[2], 2)
            elif type(x[2]) == list:
                found += self.getAtomPos(atom, x[2], l)
                l += len(self.getData(x[2], justlen=1))
            else:
                l += len(x[2])
        return found

    def delAtom(self, atom, tree=None, keepempty=1):
        if len(atom) != 4: return False
        if not tree: tl=1;tree=self.tree
        else: tl=0
        good = 0
        for x in tree:
            if x[1] == atom:
                x[3] = 1
            elif type(x[2]) == list:
                x[2] = self.delAtom(atom,x[2])
                if keepempty or len(x[2]):
                    good += 1
                else:
                    x[3] = 1
            else: good += 1
        if tl: self.tree=tree; return self
        elif keepempty or good > 0: return tree
        else: return []

    def getAtomTree(self, data=None):
        if not data: tl = 1; data = self.sbuffer
        else: tl = 0
        tree = []
        while len(data):
            size = uint32(data)
            type = data[4:8]
            if size == 1:
                size = uint64(data[8:16])
                off = 8
            elif size == 0:
                if tl:
                    size = len(data)
                else:
                    print 'Error! Cannot have an atom with a size of 0 that is not top-level.'
                    return False
                off = 0
            else: off = 0
            atom = data[off + 8:size]
            self.i += 1
            if type in self.containers:
                tree.append([self.i - 1, type, self.getAtomTree(atom), 0])
            elif type in self.special:
                tree.append([self.i - 1, type, self.special[type](self, atom), 0])
            else:
                tree.append([self.i - 1, type, atom, 0])
            data = data[size:]
        return tree
