"""
  PyMusique - iTMS client

  Travis Watkins <alleykat@gmail.com>
  Cody Brocious <daeken@gmail.com>
  Jon Lech Johansen <jon@nanocrew.net>
  Charles M. Hannum <root@ihack.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  (C) Copyright 2005 Travis Watkins, Cody Brocious, Jon Lech Johansen,
  Charles M. Hannum
"""

import os
import re
import sys
import md5
import random
import xml.parsers.expat
import config
from binascii import a2b_hex
from twisted.web.client import _parse, HTTPClientFactory, reactor, HTTPPageGetter

from scramble import *
import mcrypt
from StringIO import StringIO
from gzip import GzipFile

from debug import *

keys = {'2':a2b_hex('8a9dad399fb014c131be611820d78895'),
        '3':a2b_hex('98d1e22caa0d8abd65d143ad275a5ee7')}
keyv = a2b_hex('44e92301a2a2778114179906b53367dd')
useragent = 'iTunes/4.7.1 (Macintosh; N; PPC Mac OS X 10.3)'

def DictParse(data, dictname=None):
    xp = XMLHandler(data)
    if dictname and dictname in xp.dict:
        return xp.dict[dictname]
    else:
        return xp.dict

class XMLHandler:
    def __init__(self, data):
        self.p = xml.parsers.expat.ParserCreate('UTF-8')
        self.p.StartElementHandler = self.start_element
        self.p.EndElementHandler = self.end_element
        self.p.CharacterDataHandler = self.char_data
        self.dict = {}
        self.mapping = []
        self.types = []
        self.need = None
        self.data = {}
        self.i = 0

        self.p.Parse(data)

    def start_element(self, name, attrs):
        if not name in ('key', 'string', 'integer', 'int', 'true', 'false', 'array', 'dict'): return
        if name in ('key', 'string', 'integer', 'int', 'true', 'false'):
            self.need = name
            self.data[name] = ''
        elif name == 'array':
            type, val = 'dict', self.dict
            for i in xrange(len(self.mapping)):
                map, type = self.mapping[i], self.types[i]
                val = val[map]
            if type == 'dict':
                self.mapping.append(self.data['key'])
                self.types.append('array')
                val[self.data['key']] = []
            else:
                self.mapping.append(len(val))
                self.types.append('array')
                val.append([])
        elif name == 'dict':
            if len(self.mapping) == 0 and not 'key' in self.data: return
            type, val = 'dict', self.dict
            for i in xrange(len(self.mapping)):
                map, type = self.mapping[i], self.types[i]
                val = val[map]
            if type == 'dict':
                self.mapping.append(self.data['key'])
                self.types.append('dict')
                val[self.data['key']] = {}
            else:
                self.mapping.append(len(val))
                self.types.append('dict')
                val.append({})

    def end_element(self, name):
        if not name in ('key', 'string', 'integer', 'int', 'true', 'false', 'array', 'dict'): return
        if name in ('dict', 'array'):
            self.mapping, self.types = self.mapping[:-1], self.types[:-1]
        type, val = 'dict', self.dict
        for i in xrange(len(self.mapping)):
            map, type = self.mapping[i], self.types[i]
            val = val[map]
        if name in ('string', 'integer', 'int'):
            if type == 'dict' and 'key' in self.data:
                val[self.data['key'].strip()] = self.data[name].strip()
            else:
                val.append(self.data[name].strip())
        self.need = None

    def char_data(self, data):
        if not len(data) or not self.need: return
        self.data[self.need] += data.encode('UTF-8')

class Progress:
    size, pos, bs = 0, 0, 1024

    def SetSize(self, size):
        self.size = size

    def Update(self, pos):
        self.pos = pos
        self.Display()

    def Display(self):
        pass

class HTTPRequest(HTTPPageGetter):
    total_length = None
    num_bytes = 0
    def rawDataReceived(self, data):
        self.num_bytes += len(data)
        if self.progress:
            if self.total_length is None:
                self.total_length = self.length
                self.progress.SetSize(self.total_length)
            self.progress.Update(self.num_bytes)
        HTTPPageGetter.rawDataReceived(self, data)

class HTTPRequestFactory(HTTPClientFactory):
    protocol = HTTPRequest
    def __init__(self, url, progress=None, headers=None, postdata=None):
        self.progress = progress
        if not postdata:
            HTTPClientFactory.__init__(self, url, headers=headers, agent=useragent)
        else:
            HTTPClientFactory.__init__(self, url, headers=headers, agent=useragent, method='POST', postdata=postdata)

    def buildProtocol(self, addr):
        p = HTTPClientFactory.buildProtocol(self, addr)
        p.progress = self.progress
        return p

class StoreBase:
    token = None
    dsid = None
    country = None

    def __init__(self):
        self.getUrlBag()

        if not os.path.isdir(config.drmsdir):
            os.mkdir(config.drmsdir)
        guidfile = os.path.join(config.drmsdir, 'GUID')
        if os.path.isfile(guidfile):
            self.guid = file(guidfile, 'r').readline().strip()
        else:
            self.guid = ''
            for i in xrange(7):
                if i > 0:
                    self.guid += '.'
                self.guid += '%04X%04X' % (random.randint(0, 65535), random.randint(0, 65535))
            file(guidfile, 'w').write(self.guid)

        self.countries = {
            'Austria':'143445-2',
            'Belgium':'143446-2',
            'Canada':'143455-6',
            'Finland':'143447',
            'France':'143442',
            'Germany':'143443',
            'Great Britain':'143444',
            'Greece':'143448',
            'Ireland':'143449',
            'Italy':'143450',
            'Luxembourg':'143451-2',
            'Netherlands':'143452',
            'Portugal':'143453',
            'Spain':'143454',
            'United States':'143441'
        }

    def getUrlBag(self):
        if not os.path.isdir(config.cachedir):
            os.mkdir(config.cachedir)

        filename = os.path.join(config.cachedir, 'storeBag.xml')
        try:
            data = file(filename, 'r').read()
            self.storeBag = DictParse(data, 'urlBag')
        except IOError:
            self.Request('http://phobos.apple.com/storeBag.xml', self.gotStoreBag)

        filename = os.path.join(config.cachedir, 'secureBag.xml')
        try:
            data = file(filename, 'r').read()
            self.secureBag = DictParse(data, 'urlBag')
        except IOError:
            self.Request('https://phobos.apple.com/secureBag.xml', self.gotSecureBag)

    def gotStoreBag(self, data):
        filename = os.path.join(config.cachedir, 'storeBag.xml')
        file(filename, 'w').write(data[1])
        self.storeBag = DictParse(data[1], 'urlBag')

    def gotSecureBag(self, data):
        filename = os.path.join(config.cachedir, 'secureBag.xml')
        file(filename, 'w').write(data[1])
        self.secureBag = DictParse(data[1], 'urlBag')

    def url(self, url, params):
        if url[:5] != 'http:' and url[:6] != 'https:':
            try:
                url = self.secureBag[url]
            except KeyError:
                url = self.storeBag[url]
        if len(params):
            if params[0] == '?':
                params = params[1:]
            url += '?' + params
        return url

    def ComputeValidator(self, url, useragent):
        rand = '%04X%04X' % (random.randint(0, 65535), random.randint(0, 65535))
        m = md5.new('/' + '/'.join(url.split('/')[3:]))
        m.update(useragent + keyv + rand)
        return rand + '-' + m.hexdigest().upper()

    def Request(self, url, callback, params = '', cookie = '', progress = None, postdata = None):
        url = self.url(url, params)
        #print "url=", url
        #if len(cookie):
            #print "cookie=", cookie
        Debug(INFO, 'store', url)

        headers = {}
        if cookie:
            headers['Cookie'] = cookie
        headers['Accept'] = '*/*'
        headers['Connection'] = 'close'
        headers['Accept-Language'] = 'en-us, en;q=0.50'
        if self.token:
            headers['X-Token'] = self.token
        if self.dsid:
            headers['X-Dsid'] = self.dsid
        headers['Accept-Encoding'] = 'gzip, x-aes-cbc'
        headers['X-Apple-Validation'] = self.ComputeValidator(url, useragent)
        if self.country:
            headers['X-Apple-Store-Front'] = self.countries[self.country]

        scheme, host, port, path = _parse(url)
        factory = HTTPRequestFactory(url, progress=progress, headers=headers, postdata=postdata)
        if scheme == 'https':
            from twisted.internet import ssl
            contextFactory = ssl.ClientContextFactory()
            reactor.connectSSL(host, port, factory, contextFactory)
        else:
            reactor.connectTCP(host, port, factory)
        factory.deferred.addCallbacks(
            callback = self.EndRequest,
            callbackArgs = (factory, callback),
            errback = self.Error)

    def Error(self, error):
        Debug(ERROR, 'store', str(error))

    def EndRequest(self, data, factory, callback):
        Debug(INFO, 'store', 'Callback got data.')
        headers = factory.response_headers

        if 'content-length' in headers:
            if len(data) != int(headers['content-length'][0]):
                print "WARNING: content length mismatch"

        if not 'content-encoding' in headers:
            callback((headers, data))
            return

        encoding = headers['content-encoding'][0].split(', ')

        sz = len(data)
        sz -= sz % 16
        if 'x-aes-cbc' in encoding:
            if 'x-apple-twofish-key' in headers:
                key = a2b_hex(headers['x-apple-twofish-key'][0])
                key = scramble(key)
            else:
                key = keys[headers['x-apple-protocol-key'][0]]
            iv = a2b_hex(headers['x-apple-crypto-iv'][0])
            m = mcrypt.MCRYPT('rijndael-128', 'cbc')
            m.init(key, iv)
            data = m.decrypt(data[:sz]) + data[sz:]
            trailer = ord(data[-1])
            if trailer:
                data = data[:-trailer]

        if 'gzip' in encoding or 'x-gzip' in encoding:
            data = GzipFile(mode='r', fileobj=StringIO(data)).read(-1)
        callback((headers, data))
